#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dicom_editor.py
"""
Classes used to edit pydicom datasets
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

from os.path import basename, dirname
from pubsub import pub
import pydicom
from pydicom.datadict import keyword_dict, get_entry
from pydicom._dicom_dict import DicomDictionary
from pydicom.errors import InvalidDicomError
from dvhaedit.utilities import remove_non_alphanumeric, get_sorted_indices


keyword_dict.pop('')  # remove the empty keyword


class DICOMEditor:
    """DICOM editing and value getter class"""
    def __init__(self, dcm, force=False):
        """
        :param dcm: either a file_path to a DICOM file or a pydicom FileDataset
        """
        if type(dcm) is not pydicom.dataset.FileDataset:
            try:
                self.dcm = pydicom.read_file(dcm, force=force)
            except InvalidDicomError:
                self.dcm = None
            self.validate_ds()

        else:
            self.dcm = dcm

        self.init_tag_values = {}
        self.history = []
        self.referenced_mode = False

        self.output_path = None

    def validate_ds(self):
        """Check for required properties in the case of an InvalidDicomError"""
        required_keywords = ['StudyDate', 'StudyTime', 'PatientID', 'StudyID', 'SeriesNumber']
        if not all([hasattr(self.dcm, keyword) for keyword in required_keywords]):
            self.dcm = None

    def edit_tag(self, new_value, tag, address=None):
        """
        Change a DICOM tag value
        :param tag: the DICOM tag of interest
        :type tag: Tag
        :param new_value: new value of the DICOM tag
        :param address: an address is required for tags within sequences
        """
        if address is None:
            old_value = self.dcm[tag].value
            self.init_tag_values[tag] = old_value
            self.dcm[tag].value = new_value
            address = [tag, old_value]
        else:
            element = self.get_element(tag, address)
            old_value = element.value
            element.value = new_value

        self.append_history(new_value, address)

        return old_value, self.get_tag_value(tag, address)

    def append_history(self, new_value, address):

        if type(new_value) is list:
            str_values = [str(v) for v in new_value]
            new_value_str = "[%s]" % ', '.join(str_values)
        else:
            new_value_str = str(new_value)

        if ',' in new_value_str:
            new_value_str = "\"%s\"" % new_value_str

        line = ','.join([str(len(self.history)+1),
                         self.address_to_string(address),
                         str(new_value_str)])
        self.history.append(line)

    @staticmethod
    def address_to_string(address):
        line = []
        for item in address[:-1]:
            tag, index = item[0], item[1]
            keyword = DicomDictionary[tag][4]
            line.append("%s[%s]" % (keyword, index))
        tag, value = tuple(address[-1])
        keyword = DicomDictionary[tag][4]

        value_str = str(value)
        if ',' in value_str:
            value = "\"%s\"" % value_str

        line.append("%s,%s" % (keyword, value))
        return '.'.join(line)

    def sync_referenced_tag(self, keyword, old_value, new_value, check_all_tags=False):
        """
        Check if there is a Referenced tag with matching value, then set to new_value if so
        :param keyword: DICOM tag keyword
        :type keyword: str
        :param old_value: if Referenced+keyword tag value is this value, update to new_value
        :param new_value: new value of tag if connected
        :param check_all_tags: Set to True to check every tag in the dataset,
        otherwise only SQ and tags with Referenced in their keywords
        :type check_all_tags: bool
        """
        if check_all_tags:
            addresses = self.find_all_tags_with_value(old_value, vr='UI')
        else:
            tag = keyword_dict.get("Referenced%s" % keyword)
            addresses = self.find_tag(tag, referenced_mode=True)

        for address in addresses:
            tag = address[-1][0]
            if self.get_tag_value(tag, address=address) == old_value:
                self.edit_tag(new_value, tag, address=address)

    def get_tag_value(self, tag, address=None):
        """
        Get the current value of the provided DICOM tag
        :param tag: the DICOM tag of interest
        :type tag: Tag
        :param address: if tag is within a sequence, an address is needed which is a list of [tag, index]
        :type address: list
        """
        return self.get_element(tag, address).value

    def get_all_tag_values(self, tag):
        all_tag_values = []
        for address in self.find_tag(tag):
            v = address[-1][1]
            all_tag_values.append("%s(%s)" % (type(v), str(v)))

        return list(set(all_tag_values))

    def get_element(self, tag, address=None):
        """
        Get the element of the provided DICOM tag
        :param tag: the DICOM tag of interest
        :type tag: Tag
        :param address: if tag is within a sequence, an address is needed which is a list of [tag, index]
        :type address: list
        """
        if address is None:
            return self.dcm[tag]

        element = self.dcm
        for sequence in address[:-1]:  # Final item in address is [tag, None]
            tag_, index_ = tuple(sequence)
            element = element[tag_][index_]
        return element[tag]

    def get_tag_keyword(self, tag):
        """
        Get the DICOM tag attribute (name)
        :param tag: the DICOM tag of interest
        :type tag: Tag
        """
        return self.dcm[tag].keyword

    def get_tag_type(self, tag):
        """
        Get the DICOM tag data type
        :param tag: the DICOM tag of interest
        :type tag: Tag
        """
        return str(type(self.dcm[tag].value)).split("'")[1]

    def save_to_file(self, file_path=None):
        """
        Save the dataset to a DICOM file with pydicom
        :param file_path: absolute file path
        :type file_path: str
        """
        file_path = self.output_path if file_path is None else file_path
        self.dcm.save_as(file_path)

    @property
    def modality(self):
        """
        Get the DICOM file type (Modality)
        :return: DICOM file modality
        :rtype: str
        """
        try:
            return str(self.dcm.Modality)
        except Exception:
            return 'Not Found'

    def find_all_tags_with_vr(self, vr):
        return self.find_tag(None, vr=vr)

    def find_all_tags_with_value(self, value, vr=None):
        return self.find_tag(None, value=value, vr=vr)

    def find_tag(self, tag, vr=None, referenced_mode=False, value=None):
        """Find all instances of tag in the pydicom dataset, return tags and indices pointing to input tag"""
        # address is a list of all values for tag, with its location
        # each item in the list has a length equal to number of tags required to identify the value
        # Example:
        #   BeamMeterSet (300A, 0086) for RT PLan is accessed via FractionGroupSequence -> ReferencedBeamSequence
        #   Therefore, each row in addresses will be:
        #     [[<FractionGroupSequence tag>, index], [<ReferencedBeamSequence tag>, index], [<BeamMeterSet tag>, value]]
        # Addresses store the int representation of tag
        #
        # To find all tags with a specified VR, set vr and set tag to None

        addresses = []
        parent = []
        vr = 'UI' if referenced_mode else vr
        self.referenced_mode = referenced_mode
        self._find_tag_instances(tag, self.dcm, addresses, parent, vr=vr, value=value)
        return addresses

    def _find_tag_instances(self, tag, data_set, addresses, parent, vr=None, value=None):
        """recursively walk through data_set sequences, collect addresses with the provided tag"""
        for elem in data_set:
            if hasattr(elem, 'VR') and hasattr(elem, 'keyword') and hasattr(elem, 'tag'):
                if not self.referenced_mode or 'Referenced' in elem.keyword:
                    if elem.VR == 'SQ':
                        for i, seq_item in enumerate(elem):
                            new_parent = parent + [[int(elem.tag), i]]
                            self._find_tag_instances(tag, seq_item, addresses, new_parent)
                    elif tag is None or elem.tag == tag or \
                            (self.referenced_mode and 'Referenced' in elem.keyword):
                        if (vr is None or vr == elem.VR) and \
                                (value is None or (elem.VR == vr and elem.value == value)):
                            v = elem.value if hasattr(elem, 'value') else None
                            address = parent + [[int(elem.tag), v]]
                            addresses.append(address)


class Tag:
    """Convert group and element strings into a keyword/tag for pydicom"""
    def __init__(self, group, element):
        """
        :param group: first parameter in a DICOM tag
        :type group: str
        :param element: second parameter in a DICOM tag
        :type element: str
        """
        self.group = self.process_string(group)
        self.element = self.process_string(element)

    def __str__(self):
        """Override str operator to return a representation for GUI"""
        return "(%s, %s)" % (self.group, self.element)

    @property
    def tag(self):
        """Get keyword/tag suitable for pydicom"""
        return tuple(['0x%s' % v for v in [self.group, self.element]])

    @property
    def tag_as_int(self):
        """Convert DICOM tag to its integer representation"""
        if self.has_x:
            return None
        string = "0x%s%s" % (self.group, self.element)
        return int(string, 16)

    @property
    def has_x(self):
        """Some retired Tags from pydicom.datadict may have an x placeholder"""
        return 'X' in self.group or 'X' in self.element

    @staticmethod
    def process_string(string):
        """
        Clean a string for tag generation
        :param string: group or element string
        :type string: str
        :return: processed string
        :rtype: str
        """
        if string.startswith('0x'):
            string = string[2:]
        return remove_non_alphanumeric(string).zfill(4).upper()

    #################################################################################
    # DICOM property getters
    #################################################################################
    @property
    def vr(self):
        return self.get_entry('VR')

    @property
    def vm(self):
        return self.get_entry('VM')

    @property
    def name(self):
        return self.get_entry('name')

    @property
    def is_retired(self):
        return self.get_entry('is_retired')

    @property
    def keyword(self):
        return self.get_entry('keyword')

    def get_entry(self, tag_property):
        """General function for the getters in the code block"""
        index = ['VR', 'VM', 'name', 'is_retired', 'keyword'].index(tag_property)
        if not self.has_x and self.group and self.element:
            try:
                return get_entry(self.tag_as_int)[index]
            except KeyError:
                pass
        return 'Not Found'


class TagSearch:
    """Class used to find partial tag keyword matches"""
    def __init__(self):
        self.keywords = list(keyword_dict)
        self.lower_case_map = {key.lower(): key for key in self.keywords}

    def __call__(self, search_str):
        return self.get_table_data(search_str)

    def get_table_data(self, search_str):
        """Return data for the ListCtrl in the main application"""
        columns = ['Keyword', 'Tag', 'VR']

        data = [(en[4], self.int_to_tag(tg), en[0]) for tg, en in self.get_matches(search_str).items() if en[4]]

        keywords = [row[0] for row in data]
        sorted_indices = get_sorted_indices(keywords)

        keywords = [data[i][0] for i in sorted_indices]
        tags = [data[i][1] for i in sorted_indices]
        value_reps = [data[i][2] for i in sorted_indices]

        data = {'Keyword': keywords, 'Tag': tags, 'VR': value_reps}
        return {'data': data, 'columns': columns}

    def get_matches(self, search_str):
        """
        Get matching tags in a pydicom._dicom_dict.DicomDictionary format
        :param search_str: a string for partial keyword of hex-tag match
        :return: partial matches
        :rtype: dict
        """
        if search_str:
            search_str = remove_non_alphanumeric(search_str).lower()
            return {tag: entry for tag, entry in DicomDictionary.items()
                    if search_str in entry[4].lower() or  # keyword match
                    search_str in remove_non_alphanumeric(str(self.int_to_tag(tag)))}  # hex tag match
        else:
            return DicomDictionary

    @staticmethod
    def keyword_to_tag(keyword):
        """
        Get a dvhaedit Tag from DICOM keyword
        :param keyword: DICOM keyword
        :type keyword: str
        :return: dvhaedit Tag class object
        :rtype: Tag
        """
        tag = str(hex(keyword_dict.get(keyword)))
        group = tag[0:-4]
        element = tag[-4:]
        return Tag(group, element)

    @staticmethod
    def hex_to_tag(tag_as_hex):
        """Convert a hex tag to Tag"""
        tag = str(tag_as_hex).zfill(8)
        group = tag[0:-4]
        element = tag[-4:]
        return Tag(group, element)

    def int_to_tag(self, tag_as_int):
        return self.hex_to_tag(str(hex(tag_as_int))[2:])

    @staticmethod
    def get_value_rep(tag_as_int):
        """Get DICOM VR with an integer DICOM tag"""
        if tag_as_int is not None:
            return get_entry(tag_as_int)[0]
        return 'Unknown'


def save_dicom(data_set):
    """Helper function for the Save Worker/Thread"""
    data_set.save_to_file()


def apply_edits(values_dicts, all_row_data, data_sets):
    """Apply the tag edits to every file in self.ds, return any errors"""
    error_log, history = [], []
    for row in range(len(all_row_data)):

        row_data = all_row_data[row]
        tag = row_data['tag']
        value_str = row_data['value_str']
        keyword = row_data['keyword']
        value_type = row_data['value_type']
        values_dict = values_dicts[row]

        for i, (file_path, ds) in enumerate(data_sets.items()):
            label = "Editing %s for file %s of %s" % (keyword, i+1, len(data_sets))
            msg = {'label': label, 'gauge': float(i) / len(data_sets)}
            pub.sendMessage("progress_update", msg=msg)
            try:
                if tag.tag in ds.dcm.keys():  # Tag exists in top-level of DICOM dataset

                    current_value = values_dict[file_path][0]
                    new_value = process_value(current_value, value_type)  # converts to list and types, as appropriate

                    old_value, _ = ds.edit_tag(new_value, tag=tag.tag)
                    history.append([keyword, old_value, new_value])
                else:  # Search entire DICOM dataset for tag
                    addresses = ds.find_tag(tag.tag)
                    if not addresses:
                        raise Exception  # Tag could not be found
                    for a, address in enumerate(addresses):
                        current_value = values_dict[file_path][a]
                        new_value = process_value(current_value, value_type)  # converts to list and types

                        old_value, _ = ds.edit_tag(new_value, tag=tag.tag, address=address)
                        history.append([keyword, old_value, new_value])

            except Exception as e:
                err_msg = 'KeyError: %s is not accessible' % tag if str(e).upper() == str(tag).upper() else e
                value = value_str if value_str else '[empty value]'
                modality = ds.dcm.Modality if hasattr(ds.dcm, 'Modality') else 'Unknown'
                error_log.append("Directory: %s\nFile: %s\nModality: %s\n\t"
                                 "Attempt to edit %s to new value: %s\n\t%s\n" %
                                 (dirname(file_path), basename(file_path), modality, tag, value, err_msg))

    return {'error_log': '\n'.join(error_log),
            'history': history,
            'ds': data_sets}


def update_referenced_tags(data_sets, check_all_tags, history_row):
    keyword, old_value, new_value = tuple(history_row)
    if "Referenced%s" % keyword in list(keyword_dict):
        for ds in data_sets:
            ds.sync_referenced_tag(keyword, old_value, new_value, check_all_tags=check_all_tags)


def is_value_list(value):
    return value[0] == '[' and value[-1] == ']'


def value_to_list(value, value_type):
    ans = []
    for v in value[1:-1].split(', '):
        if v[0] == "'" and v[-1] == "'":
            ans.append(value_type(value[1:-1]))
        else:
            ans.append(value_type(v))
    return ans


def process_value(value, value_type):
    if is_value_list(value):
        return value_to_list(value, value_type)
    return value_type(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dynamic_value.py
"""
Apply dynamic value functions
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

from os import sep
from os.path import normpath, splitext
from pubsub import pub
from pydicom.uid import generate_uid
from secrets import randbelow
from dvhaedit.paths import DYNAMIC_VALUE_HELP


class ValueGenerator:
    """Process function calls in value string"""
    def __init__(self, value=None, tag=None, options=None):
        """
        :param value: input tag value from main frame
        :type value: str
        :param tag: the tag to be edited
        """
        self.value = value
        self.tag = tag
        self.options = options
        self.enum_instances = {'file': {}, 'value': []}
        self.uids = {'file': {}, 'value': {}}
        self.rand = {'file': {}, 'value': {}}
        self.data_sets = {}
        self.file_paths = []
        self.func_call = []

        self.functions = ['file', 'val', 'fenum', 'venum', 'fuid', 'vuid', 'frand', 'vrand']
        self.value_functions = [f for f in self.functions if f[0] == 'v']
        self.func_map = {f: getattr(self, f) for f in self.functions}

        if value is not None and tag is not None:
            self.set_func_call_dict()

    def __call__(self, data_sets, file_path=None, callback=None):
        """
        :param data_sets: parsed dicom data using DICOMEditor class
        :type data_sets: dict
        :param file_path: optionally specify a single file_path for the generator to use (for efficient previewing)
        :type file_path: str
        :param callback: optional function call through each loop through files, two parameters (iteration, count_total)
        :return: new tag values
        :rtype: dict or list
        """
        if file_path is None:
            self.file_paths = sorted(list(data_sets))
        else:
            self.file_paths = [file_path]
        self.data_sets = [data_sets[f] for f in self.file_paths]

        file_count = len(self.file_paths)
        self.set_enum_instances()
        new_values = {}

        self.value_generator_callback(0, file_count)

        for f_counter, f in enumerate(self.file_paths):
            addresses = self.data_sets[f_counter].find_tag(self.tag)
            new_values[f] = [self.value.split('*') for _ in range(len(addresses))]
            for i, call_str in enumerate(self.value.split('*')):
                if i % 2 == 1:
                    for index, address in enumerate(addresses):
                        new_values[f][index][i] = str(self.get_value(call_str, f, index=index))
            new_values[f] = [''.join(new_values) for new_values in new_values[f]]

            self.value_generator_callback(f_counter+1, file_count)

        if file_path is not None:
            return new_values[file_path]
        return new_values

    @staticmethod
    def value_generator_callback(iteration, count_total):
        msg = {'label': 'Calculating values for file %s of %s' % (iteration, count_total),
               'gauge': iteration / count_total}
        pub.sendMessage("progress_update", msg=msg)

    #################################################################################
    # Setters
    #################################################################################
    def set_func_call_dict(self):
        """Create a dict of functions to input"""
        self.func_call = []
        for f in str(self.value).split('*')[1::2]:
            func, param = self.split_call_str(f)
            self.func_call.append((func, param))

    @staticmethod
    def send_progress_update(progress, label='Initializing...'):
        msg = {'label': label,
               'gauge': progress}
        pub.sendMessage("progress_update", msg=msg)

    def set_enum_instances(self):
        """Collect all unique values for each of the enumerators"""
        self.send_progress_update(0)
        # Determine how many actions to perform, for progress indication

        self.enum_instances = {'file': {}, 'value': {}}
        file_functions = ['fenum', 'fuid', 'frand']
        parameters = set(self.get_parameters(file_functions))
        count = len(parameters) + len(self.data_sets)

        # File-based instances
        counter = 0.
        for index in parameters:
            enum = [self.file(index, f, True) for f in self.file_paths]
            self.enum_instances['file'][index] = sorted(list(set(enum)))
            self.send_progress_update(0.95 * (counter / count))
            counter += 1

        # Value-based instances
        enum = []
        all_float = True
        for ds in self.data_sets:
            values = ds.get_all_tag_values(self.tag)
            for i in range(len(values)):
                try:
                    values[i] = "%0.5f" % float(values[i])
                except ValueError:
                    all_float = False
            enum.extend(values)
            self.send_progress_update(0.95 * (counter / count))
            counter += 1
        enum = list(set(enum))
        if all_float:
            enum.sort(key=float)
        self.enum_instances['value'] = enum

        # set uids
        prefix = self.options.prefix if hasattr(self.options, 'prefix') else None
        if prefix == '':
            prefix = None

        entropy_srcs = [self.options.entropy_source] if hasattr(self.options, 'entropy_source') else None
        if entropy_srcs == ['']:
            entropy_srcs = None

        self.send_progress_update(0.95, label='Generating UIDs...')
        self.uids = {'file': {}, 'value': {}}
        all_uids = []

        # File-based uids
        for index in self.get_parameters('fuid'):
            self.uids['file'][index] = {}
            for i in self.enum_instances['file'][index]:
                found, uid = False, ''
                while not found:  # ensure random number is not used twice
                    uid = generate_uid(prefix=prefix, entropy_srcs=entropy_srcs)
                    found = uid not in all_uids
                    if not found and entropy_srcs is not None:  # increment entropy source
                        entropy_srcs = [entropy_srcs[0] + '0']
                all_uids.append(uid)
                self.uids['file'][index][i] = uid

        # Value-based uids
        for i in self.enum_instances['value']:
            found, uid = False, ''
            while not found:  # ensure random number is not used twice
                uid = generate_uid(prefix=prefix, entropy_srcs=entropy_srcs)
                if uid not in all_uids:
                    found = True
                elif entropy_srcs:
                    entropy_srcs = [entropy_srcs[0] + '0']
            all_uids.append(uid)
            self.uids['value'][i] = uid

        # set random numbers
        self.send_progress_update(0.98, label='Generating random numbers...')
        digits = self.options.rand_digits if hasattr(self.options, 'rand_digits') else 5
        max_num = 10 ** digits
        self.rand = {'file': {}, 'value': {}}

        # count the number of rand_numbers needed
        count = len(self.enum_instances['file']) * len(self.get_parameters('frand')) + len(self.enum_instances['value'])

        while max_num < count:  # ensure enough digits are used, to avoid infinite while loop below
            max_num *= 10
        random_numbers = []
        for index in self.get_parameters('frand'):
            self.rand['file'][index] = {}
            for i in self.enum_instances['file'][index]:
                found, random_number = False, 0
                while not found:  # ensure random number is not used twice
                    random_number = randbelow(max_num)
                    found = random_number not in random_numbers
                random_numbers.append(random_number)
                self.rand['file'][index][i] = str(random_number).zfill(digits)

        for i in self.enum_instances['value']:
            found, random_number = False, 0
            while not found:  # ensure random number is not used twice
                random_number = randbelow(max_num)
                found = random_number not in random_numbers
            random_numbers.append(random_number)
            self.rand['value'][i] = str(random_number).zfill(digits)

    #################################################################################
    # Getters
    #################################################################################
    def get_parameters(self, functions):
        """Get a the list of parameters for the specified function"""
        parameters = []
        for call_str in self.value.split('*')[1::2]:  # every odd index
            func, param = self.split_call_str(call_str)
            if func in functions:
                parameters.append(param)
        return sorted(list(set(parameters)))

    def get_value(self, call_str, file_path, index=None):
        """Parse the function call string, perform the function"""
        func, param = self.split_call_str(call_str, index)
        if func in list(self.func_map):
            return self.func_map[func](param, file_path)
        return ''

    #################################################################################
    # Utilities
    #################################################################################
    def split_call_str(self, func_call_str, index=None):
        """Split the string into function and parameter"""
        if func_call_str in self.value_functions:
            return func_call_str, index
        f_split = func_call_str.split('[')
        func = f_split[0]
        param = f_split[1][:-1]  # remove last character, ]
        return func, int(param)

    #################################################################################
    # Functions
    #################################################################################
    @staticmethod
    def file(index, file_path, all_up_to_index=False):
        """Process a directory name"""
        components = normpath(file_path).split(sep)
        if all_up_to_index:
            if index == -1:
                return splitext('/'.join(components))[0]
            else:
                return splitext('/'.join(components[:index+1]))[0]
        return splitext(components[index])[0]

    # The following function names, in this comment block, must match those found in self.functions,
    # and have two input parameters: index and file_path

    def val(self, index, file_path):
        """Process a value enumeration (index included for abstract use)"""
        ds = self.data_sets[self.file_paths.index(file_path)]
        try:
            return ds.get_tag_value(self.tag)
        except KeyError:
            addresses = ds.find_tag(self.tag)
            if addresses:
                return addresses[index][-1][1]
            return None

    def fenum(self, index, file_path):
        """Process a file enumeration"""
        return str(self.enum_instances['file'][index].index(self.file(index, file_path, True)) + 1)

    def venum(self, index, file_path):
        """Process a value enumeration"""
        ds = self.data_sets[self.file_paths.index(file_path)]
        try:
            value = ds.get_tag_value(self.tag)
        except KeyError:
            addresses = ds.find_tag(self.tag)
            if addresses:
                value = addresses[index][-1][1]
                try:
                    value = "%0.5f" % float(addresses[index][-1][1])
                except ValueError:
                    pass
            else:
                return 'None'
        return str(self.enum_instances['value'].index(str(value)) + 1)

    def fuid(self, index, file_path):
        """Process file path based uid generator"""
        return self.fmethod(index, file_path, self.uids)

    def vuid(self, index, file_path):
        """Process value based uid generator"""
        return self.vmethod(index, file_path, self.uids)

    def frand(self, index, file_path):
        """Process file path based random number generator"""
        return self.fmethod(index, file_path, self.rand)

    def vrand(self, index, file_path):
        """Process value based random number generator"""
        return self.vmethod(index, file_path, self.rand)

    # Helper functions for file and value type functions above, to reduce repeated code
    def fmethod(self, index, file_path, lookup):
        """Process a file-like function (except enum)"""
        value = self.file(index, file_path, True)
        return lookup['file'][index][value]

    def vmethod(self, index, file_path, lookup):
        """Process a value-like function (except enum)"""
        ds = self.data_sets[self.file_paths.index(file_path)]
        try:
            value = ds.get_tag_value(self.tag)
        except KeyError:
            addresses = ds.find_tag(self.tag)
            if addresses:
                value = addresses[index][-1][1]
            else:
                return None
        return lookup['value'][str(value)]


with open(DYNAMIC_VALUE_HELP, 'r') as doc:
    HELP_TEXT = doc.read()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# main.py
"""
The main file for DVHA DICOM Editor
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

import wx
from copy import deepcopy
from datetime import datetime
from os import sep
from os.path import isdir, isfile, basename, join, dirname, normpath, splitext, relpath
from pathlib import Path
from pubsub import pub
from pydicom.datadict import keyword_dict
import webbrowser
from dvhaedit._version import __version__
from dvhaedit.data_table import DataTable
from dvhaedit.dialogs import ErrorDialog, ViewErrorLog, AskYesNo, TagSearchDialog, About, DynamicValueHelp,\
    AdvancedSettings
from dvhaedit.threads import ParsingProgressFrame, SavingProgressFrame, RefSyncProgressFrame, ValueGenProgressFrame,\
    ApplyEditsProgressFrame
from dvhaedit.dicom_editor import Tag
from dvhaedit.dynamic_value import ValueGenerator
from dvhaedit.options import Options
from dvhaedit.utilities import set_msw_background_color, get_file_paths, get_type, get_selected_listctrl_items,\
    is_mac, save_object_to_file, load_object_from_file, set_frame_icon


class MainFrame(wx.Frame):
    """The main frame called in MainApp"""
    def __init__(self, *args, **kwds):
        kwds["style"] = kwds.get("style", 0) | wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)

        self.ds = {}
        self.functions = ValueGenerator().functions
        self.all_options = {}
        self.current_options = Options()
        self.value_generators = []
        self.values_dicts = []

        self.error_log = ''
        self.history = []

        # Create GUI widgets
        keys = ['in_dir', 'tag_group', 'tag_element', 'value', 'out_dir', 'prepend_file_name']
        self.input = {key: wx.TextCtrl(self, wx.ID_ANY, "") for key in keys}
        self.input['preview'] = wx.TextCtrl(self, wx.ID_ANY, "", style=wx.TE_READONLY)
        self.input['selected_file'] = wx.ComboBox(self, wx.ID_ANY, style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.input['value_type'] = wx.ComboBox(self, wx.ID_ANY, style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.input_text_obj = [self.input[key] for key in keys]  # use for text event binding and focusing

        keys = ['save_dicom', 'quit', 'in_browse', 'out_browse', 'add', 'delete', 'select_all', 'deselect_all',
                'save_template', 'load_template', 'advanced']
        self.button = {key: wx.Button(self, wx.ID_ANY, key.replace('_', ' ').title()) for key in keys}
        bmp = wx.ArtProvider.GetBitmap(wx.ART_FIND, size=(16, 16))
        self.button['search'] = wx.BitmapButton(self, id=wx.ID_ANY, bitmap=bmp)
        bmp = wx.ArtProvider.GetBitmap(wx.ART_INFORMATION, size=(16, 16))
        self.button['value_help'] = wx.BitmapButton(self, id=wx.ID_ANY, bitmap=bmp)

        columns = ['Index', 'Tag', 'Keyword', 'Value', 'Value Type']
        data = {c: [''] for c in columns}
        self.list_ctrl = wx.ListCtrl(self, wx.ID_ANY, style=wx.BORDER_SUNKEN | wx.LC_REPORT)
        self.data_table = DataTable(self.list_ctrl, data=data, columns=columns, widths=[-2] * 5)

        keys = ['tag_group', 'tag_element', 'value', 'value_type', 'files_found', 'keyword', 'selected_file',
                'modality', 'prepend_file_name', 'search', 'value_rep', 'preview', 'advanced', 'value_help']
        self.label = {key: wx.StaticText(self, wx.ID_ANY, key.replace('_', ' ').title() + ':') for key in keys}

        self.search_sub_folders = wx.CheckBox(self, wx.ID_ANY, "Search Sub-Folders")
        self.search_sub_folders_last_status = False

        self.force_open = wx.CheckBox(self, wx.ID_ANY, "Force DICOM Read")

        self.retain_rel_dir = wx.CheckBox(self, wx.ID_ANY, "Retain relative directory structure")
        self.save_history = wx.CheckBox(self, wx.ID_ANY, "Save edit log")

        self.referenced_tag_choices = ['Only Edit Tags Defined in Table',
                                       'Update "Referenced" Tags',
                                       'Update All Tags with Matching UID']
        self.update_referenced_tags = wx.ComboBox(self, wx.ID_ANY, style=wx.CB_DROPDOWN | wx.CB_READONLY,
                                                  choices=self.referenced_tag_choices)

        self.file_paths = []
        self.ignored_file_paths = []
        self.update_files_found()
        self.refresh_needed = False

        self.directory = {'in': '', 'out': ''}

        self.__set_properties()
        self.__add_menubar()
        self.__do_subscribe()
        self.__do_bind()
        self.__do_layout()

    def __set_properties(self):
        """Set initial properties of widgets"""
        set_msw_background_color(self)

        for checkbox in [self.search_sub_folders, self.retain_rel_dir, self.save_history, self.force_open]:
            checkbox.SetFont(wx.Font(11, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, 0, ""))
        self.retain_rel_dir.SetToolTip("If unchecked, all new files will be placed in the output directory. "
                                       "Otherwise, the same relative directory structure will be used.")
        self.retain_rel_dir.SetValue(True)
        self.save_history.SetToolTip("Save a text file of tag edits to the output directory.")
        self.save_history.SetValue(True)

        self.button['in_browse'].SetLabel(u"Browse…")
        self.button['out_browse'].SetLabel(u"Browse…")
        self.button['save_dicom'].SetLabel('Save DICOM')
        self.button['save_template'].SetLabel('Save')
        self.button['load_template'].SetLabel('Load')

        self.label['advanced'].SetLabel(' ')
        self.label['search'].SetLabel(' ')
        self.label['value_help'].SetLabel(' ')
        self.label['value_rep'].SetLabel('Value Representation (VR): ')
        self.label['value'].SetLabel('Enter New DICOM Tag Value Here: ')

        self.button['search'].SetToolTip("Search for DICOM tag based on keyword.")
        self.input['preview'].SetToolTip("Values may be set dynamically, a preview is shown here. Note that generated "
                                         "UIDs may be different than the final value if no entropy source is "
                                         "provided.")
        if is_mac():
            self.input['preview'].SetBackgroundColour((230, 230, 230))

        for key in ['add', 'delete', 'save_dicom', 'save_template', 'advanced']:
            self.button[key].Disable()

        self.input['selected_file'].Disable()
        msg = "The \"Value\" in the tag editor below will be prepopulated with the value found for the " \
              "specified tag in this file."
        self.input['selected_file'].SetToolTip(msg)
        self.label['selected_file'].SetToolTip(msg)

        self.input['value_type'].SetItems(['str', 'float', 'int'])
        self.input['value_type'].SetValue('str')

        self.update_referenced_tags.SetValue(self.referenced_tag_choices[1])
        self.update_referenced_tags.SetToolTip("Automatically sync Referenced<tag> to new <tag> value to maintain "
                                               "cross-file connections (e.g., keep RT-Structure connection to "
                                               "RT-Plan). If this doesn't work, try Update All Tags with Matching UID "
                                               "(much slower).")

    def __add_menubar(self):

        self.frame_menubar = wx.MenuBar()

        file_menu = wx.Menu()
        menu_open = file_menu.Append(wx.ID_OPEN, '&Open\tCtrl+O')
        self.menu_save = file_menu.Append(wx.ID_ANY, '&Save\tCtrl+S')
        self.menu_save.Enable(False)
        file_menu.Append(wx.ID_SEPARATOR)
        qmi = file_menu.Append(wx.ID_ANY, '&Quit\tCtrl+Q')

        help_menu = wx.Menu()
        menu_github = help_menu.Append(wx.ID_ANY, 'GitHub Page')
        menu_report_issue = help_menu.Append(wx.ID_ANY, 'Report an Issue')
        menu_about = help_menu.Append(wx.ID_ANY, '&About')

        self.Bind(wx.EVT_MENU, self.on_quit, qmi)
        self.Bind(wx.EVT_MENU, self.on_load_template, menu_open)
        self.Bind(wx.EVT_MENU, self.on_save_template, self.menu_save)
        self.Bind(wx.EVT_MENU, self.on_githubpage, menu_github)
        self.Bind(wx.EVT_MENU, self.on_report_issue, menu_report_issue)
        file_menu.Append(wx.ID_SEPARATOR)
        self.Bind(wx.EVT_MENU, self.on_about, menu_about)

        self.frame_menubar.Append(file_menu, '&File')
        self.frame_menubar.Append(help_menu, '&Help')
        self.SetMenuBar(self.frame_menubar)

    def __do_bind(self):
        """Bind user events to widgets with actions"""

        # This requires that there is a function on_<button-key> for every button, simplifies bind
        for key, button in self.button.items():
            self.Bind(wx.EVT_BUTTON, getattr(self, "on_" + key), id=button.GetId())

        self.Bind(wx.EVT_COMBOBOX, self.on_file_select, id=self.input['selected_file'].GetId())

        for widget in self.input_text_obj:
            widget.Bind(wx.EVT_KEY_UP, self.on_key_up)

        for key in ['in_dir', 'out_dir']:
            self.input[key].Bind(wx.EVT_KEY_DOWN, self.on_key_down_dir)
            self.input[key].Bind(wx.EVT_TEXT, self.update_dir_obj_text_color)

        self.input['value'].Bind(wx.EVT_TEXT, self.update_add_enable)

        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_selection, id=self.list_ctrl.GetId())

    def __do_subscribe(self):
        pub.subscribe(self.add_parsed_data, "add_parsed_data")
        pub.subscribe(self.on_parse_complete, "parse_complete")
        pub.subscribe(self.on_save_complete, 'save_complete')
        pub.subscribe(self.do_saving_progress_frame, 'ref_sync_complete')
        pub.subscribe(self.add_value_dicts, 'add_value_dicts')
        pub.subscribe(self.call_next_value_generator, 'value_gen_complete')
        pub.subscribe(self.update_dicom_edits, 'update_dicom_edits')
        pub.subscribe(self.do_save_dicom, 'do_save_dicom')

    def __do_layout(self):
        """Create GUI layout"""
        # Create GUI sizers
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_input_dir_wrapper = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Directory"), wx.VERTICAL)
        sizer_input_dir = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input_file = wx.BoxSizer(wx.VERTICAL)
        sizer_edit_wrapper = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Tag Editor"), wx.VERTICAL)
        sizer_edit = wx.BoxSizer(wx.HORIZONTAL)
        sizer_edit_widgets = {key: wx.BoxSizer(wx.VERTICAL)
                              for key in ['tag_group', 'tag_element', 'value', 'value_type', 'value_help',
                                          'add', 'search', 'advanced']}
        sizer_value_keyword = wx.BoxSizer(wx.VERTICAL)
        sizer_edit_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_output_dir_wrapper = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Output Directory"), wx.HORIZONTAL)
        sizer_output_dir_inner_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_output_checkboxes = wx.BoxSizer(wx.HORIZONTAL)
        sizer_output_dir = wx.BoxSizer(wx.VERTICAL)
        sizer_output_dir_prepend = wx.BoxSizer(wx.HORIZONTAL)
        sizer_app_buttons = wx.BoxSizer(wx.HORIZONTAL)

        # Directory Browser
        sizer_input_dir.Add(self.input['in_dir'], 1, wx.EXPAND | wx.ALL, 5)
        sizer_input_dir.Add(self.button['in_browse'], 0, wx.ALL, 5)
        sizer_input_dir_wrapper.Add(sizer_input_dir, 0, wx.ALL | wx.EXPAND, 5)
        row_sizer = wx.BoxSizer(wx.HORIZONTAL)
        row_sizer.Add(self.label['files_found'], 1, wx.EXPAND, 0)
        row_sizer.Add(self.force_open, 0, wx.ALIGN_RIGHT | wx.RIGHT, 15)
        row_sizer.Add(self.search_sub_folders, 0, wx.ALIGN_RIGHT, 0)
        sizer_input_file.Add(row_sizer, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5)
        sizer_input_file.Add(self.label['selected_file'], 0, wx.LEFT | wx.TOP, 5)
        sizer_input_file.Add(self.input['selected_file'], 0, wx.EXPAND | wx.LEFT | wx.BOTTOM, 5)
        sizer_input_file.Add(self.label['modality'], 0, wx.LEFT, 5)
        sizer_input_dir_wrapper.Add(sizer_input_file, 0, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_input_dir_wrapper, 0, wx.EXPAND | wx.ALL, 5)

        # Input Widgets
        for key in ['search', 'tag_group', 'tag_element', 'value_type', 'value_help', 'advanced']:
            obj = self.button if key in {'search', 'value_help', 'advanced'} else self.input
            sizer_edit_widgets[key].Add(self.label[key], 0, wx.EXPAND, 0)
            sizer_edit_widgets[key].Add(obj[key], 0, wx.EXPAND, 0)
            proportion = 'tag' in key
            sizer_edit.Add(sizer_edit_widgets[key], proportion, wx.EXPAND | wx.ALL, 5)
        sizer_edit_wrapper.Add(sizer_edit, 0, wx.EXPAND | wx.ALL, 5)

        sizer_value_keyword.Add(self.label['value'], 0, wx.LEFT, 5)
        row_sizer_value_help = wx.BoxSizer(wx.HORIZONTAL)
        row_sizer_value_help.Add(self.input['value'], 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        row_sizer_value_help.Add(self.button['add'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_value_keyword.Add(row_sizer_value_help, 1, wx.EXPAND, 5)
        sizer_value_keyword.Add(self.label['keyword'], 0, wx.TOP | wx.LEFT, 5)
        sizer_value_keyword.Add(self.label['value_rep'], 0, wx.LEFT | wx.BOTTOM, 5)
        sizer_value_keyword.Add(self.label['preview'], 0, wx.BOTTOM | wx.LEFT, 5)
        sizer_value_keyword.Add(self.input['preview'], 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_edit_wrapper.Add(sizer_value_keyword, 0, wx.EXPAND | wx.ALL, 5)

        sizer_edit_wrapper.Add(self.list_ctrl, 1, wx.EXPAND | wx.ALL, 10)
        for key in ['delete', 'select_all', 'deselect_all', 'save_template', 'load_template']:
            sizer_edit_buttons.Add(self.button[key], 0, wx.EXPAND | wx.RIGHT | wx.LEFT, 5)
        sizer_edit_wrapper.Add(sizer_edit_buttons, 0, wx.ALIGN_CENTER | wx.ALL, 5)

        sizer_main.Add(sizer_edit_wrapper, 1, wx.EXPAND | wx.ALL, 5)

        # Output Directory Browser
        sizer_output_dir.Add(self.input['out_dir'], 1, wx.EXPAND | wx.ALL, 5)
        sizer_output_checkboxes.Add(self.retain_rel_dir, 0, wx.RIGHT, 15)
        sizer_output_checkboxes.Add(self.save_history, 0, 0, 5)
        sizer_output_dir.Add(sizer_output_checkboxes, 0, wx.LEFT | wx.BOTTOM, 5)
        sizer_output_dir_inner_wrapper.Add(sizer_output_dir, 0, wx.ALL | wx.EXPAND, 5)
        sizer_output_dir_prepend.Add(self.label['prepend_file_name'], 0, wx.TOP | wx.LEFT | wx.RIGHT, 5)
        sizer_output_dir_prepend.Add(self.input['prepend_file_name'], 1, wx.EXPAND | wx.RIGHT, 5)
        sizer_output_dir_inner_wrapper.Add(sizer_output_dir_prepend, 0, wx.EXPAND | wx.LEFT | wx.BOTTOM | wx.RIGHT, 5)
        sizer_output_dir_wrapper.Add(sizer_output_dir_inner_wrapper, 1, wx.EXPAND, 5)
        sizer_output_dir_wrapper.Add(self.button['out_browse'], 0, wx.ALIGN_TOP | wx.TOP, 10)
        sizer_main.Add(sizer_output_dir_wrapper, 0, wx.EXPAND | wx.ALL, 5)

        sizer_app_buttons.Add(self.update_referenced_tags, 0, wx.ALIGN_LEFT | wx.ALL, 5)
        sizer_app_buttons.Add((20, 20), 1, wx.EXPAND | wx.ALIGN_LEFT | wx.ALL, 5)
        sizer_app_buttons.Add(self.button['save_dicom'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_app_buttons.Add(self.button['quit'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_main.Add(sizer_app_buttons, 0, wx.EXPAND | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.SetMinSize(self.GetSize())
        self.Center()

    #################################################################################
    # Basic properties/getters
    #################################################################################
    @property
    def group(self):
        """Group of the DICOM tag"""
        return self.input['tag_group'].GetValue()

    @property
    def element(self):
        """Element of the DICOM tag"""
        return self.input['tag_element'].GetValue()

    @property
    def tag(self):
        """
        Convert string input in GUI to a DICOM tag
        :return: user provided tag
        :rtype: Tag
        """
        return Tag(self.group, self.element)

    @property
    def value(self):
        value = self.input['value'].GetValue()
        type_ = get_type(self.input['value_type'].GetValue())
        if len(value) and value[0] == '[' and value[-1] == ']':
            return value
        return type_(value)

    @property
    def keyword(self):
        return self.tag.keyword

    @property
    def selected_indices(self):
        return get_selected_listctrl_items(self.list_ctrl)

    @property
    def selected_file(self):
        return self.input['selected_file'].GetSelection()

    @property
    def selected_data_set(self):
        return self.ds.get(self.file_paths[self.selected_file])

    @property
    def data_table_has_data(self):
        return self.data_table.has_data and self.data_table.get_row(0)[0] != ''

    #################################################################################
    # Button Event tickers
    # Binding occurs abstractly with a for loop, all buttons automatically bind to
    # a function of its dict key prepended with on_
    # For example, the Add button is self.button['add'], clicking it will activate
    # self.on_add
    #################################################################################
    def on_in_browse(self, *evt):
        self.on_browse(self.input['in_dir'])
        if self.refresh_needed:
            self.refresh_ds()
            self.refresh_needed = False

    def on_out_browse(self, *evt):
        self.on_browse(self.input['out_dir'])

    def on_browse(self, obj):
        """
        General function for in and out browse events
        :param obj: either in_dir or out_dir TextCtrl objects
        :type obj: wx.TextCtrl
        """
        starting_dir = obj.GetValue()
        if not isdir(starting_dir):
            starting_dir = ""

        dlg = wx.DirDialog(self, "Select directory", starting_dir, wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        if dlg.ShowModal() == wx.ID_OK:
            new_dir = dlg.GetPath()
            dlg.Destroy()
            obj.SetBackgroundColour(wx.WHITE)  # Reset if background was orange, DirDialog forces a valid directory
            obj.ChangeValue(new_dir)  # Update TextCtrl without signaling a change event
            if obj == self.input['in_dir']:
                self.refresh_needed = True
            else:
                self.directory['out'] = new_dir
            self.input['tag_group'].SetFocus()
        else:
            obj.SetFocus()
        self.update_save_dicom_enable()

    def on_add(self, *evt):
        """Add a tag edit"""
        keys_int = [int(key) for key in list(self.all_options)]
        row_index = '0' if not keys_int else str(max(keys_int) + 1)
        self.all_options[row_index] = deepcopy(self.current_options)
        row = [row_index,
               str(self.tag),
               self.keyword,
               self.input['value'].GetValue(),
               self.input['value_type'].GetValue()]
        if self.data_table_has_data:
            self.data_table.append_row(row)
        else:
            columns = self.data_table.columns
            data = {columns[i]: [value] for i, value in enumerate(row)}
            self.data_table.set_data(data, columns)
        self.data_table.set_column_widths(auto=True)

        for key in ['tag_group', 'tag_element', 'value']:
            self.input[key].SetValue('')

        self.input['tag_group'].SetFocus()
        self.update_keyword()
        self.update_save_template_enable()
        self.update_save_dicom_enable()

    def on_search(self, *evt):
        TagSearchDialog(self)

    def on_advanced(self, *evt):
        AdvancedSettings(self.current_options)
        self.update_preview()

    @staticmethod
    def on_value_help(*evt):
        DynamicValueHelp()

    def on_delete(self, *evt):
        """Delete the selected tag edits"""
        for index in self.selected_indices[::-1]:
            self.data_table.delete_row(index)
        self.update_delete_enable()
        self.update_save_template_enable()
        self.update_save_dicom_enable()

    def on_select_all(self, *evt):
        """Select all tag edits"""
        self.data_table.apply_selection_to_all(True)
        self.button['delete'].Enable()

    def on_deselect_all(self, *evt):
        """Deselect all tag edits"""
        self.data_table.apply_selection_to_all(False)
        self.button['delete'].Disable()

    def on_save_template(self, *evt):
        """Save the current tag edits to a pickle file"""
        dlg = wx.FileDialog(self, "Save template", "", wildcard='*.pickle',
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if dlg.ShowModal() == wx.ID_OK:
            save_data = {'table': self.data_table.get_save_data(), 'options': self.all_options}
            save_object_to_file(save_data, dlg.GetPath())
        dlg.Destroy()

    def on_load_template(self, *evt):
        """Load a pickle file of tag edits"""
        dlg = wx.FileDialog(self, "Load template", "", wildcard='*.pickle', style=wx.FD_OPEN)
        if dlg.ShowModal() == wx.ID_OK:
            try:
                loaded_data = load_object_from_file(dlg.GetPath())
                self.data_table.load_save_data(loaded_data['table'])
                self.all_options = loaded_data['options']
            except Exception:
                self.data_table.clear()

        self.update_save_template_enable()
        self.update_save_dicom_enable()

    def on_save_dicom(self, *evt):
        """Apply edits, check for errors, then run save_files"""
        if self.dir_contents_have_changed:
            with AskYesNo(self, "Directory contents have changed. Continue anyway?") as dlg:
                if dlg.ShowModal() == wx.ID_NO:
                    return

        # Can be expensive, on_save_dicom, split to enable threading
        # calls do_save_dicom when done
        self.calculate_value_generators()

    def on_quit(self, *evt):
        self.Close()

    #################################################################################
    # ListCtrl (Table) Event tickers
    #################################################################################
    def on_selection(self, *evt):
        self.update_delete_enable()
        selected_data = self.data_table.selected_row_data
        if selected_data:
            tag = selected_data[0][1][1:-1].split(',')
            group = tag[0].strip()
            element = tag[1].strip()
            self.input['tag_group'].SetValue(group)
            self.input['tag_element'].SetValue(element)
            self.input['value_type'].SetValue(selected_data[0][4])
            self.update_keyword()
            self.update_init_value()
            self.update_preview()
            self.current_options = deepcopy(self.all_options[selected_data[0][0]])

    #################################################################################
    # Key Event tickers
    #################################################################################
    def on_key_up(self, evt):
        """Called anytime a user's key is released"""
        keycode = evt.GetKeyCode()
        if keycode == wx.WXK_TAB:
            self.on_tab_key(evt)
        elif evt.GetEventObject() in {self.input['tag_group'], self.input['tag_element']}:
            self.update_keyword()
        evt.Skip()

    def on_tab_key(self, evt):
        obj = evt.GetEventObject()
        index = self.input_text_obj.index(obj)
        index = index + 1 if index + 1 < len(self.input_text_obj) else 0
        if obj in {self.input['in_dir'], self.input['out_dir']}:
            new_dir = obj.GetValue()
            if isdir(new_dir):
                if obj == self.input['in_dir']:
                    self.refresh_ds()
            else:
                ErrorDialog(self, "Please enter a valid directory.", "Directory Error")
                dir_key = 'in' if obj == self.input['in_dir'] else 'out'
                self.directory[dir_key] = new_dir
                if obj == self.input['in_dir']:
                    self.file_paths = []
                    self.ds = {}
                    self.on_parse_complete()
                index -= 1
        self.input_text_obj[index].SetFocus()
        self.update_save_dicom_enable()
        self.update_keyword()
        if obj == self.input['in_dir']:
            self.update_init_value()

    def on_key_down_dir(self, evt):
        """Called on any key down in directory TextCtrl, only act on Enter/Return"""
        keycode = evt.GetKeyCode()
        obj = evt.GetEventObject()
        if keycode == wx.WXK_RETURN:
            self.on_enter_key_dir(obj)
            if obj == self.input['in_dir']:
                self.update_keyword()
                self.update_init_value()
        else:
            evt.Skip()

    def on_enter_key_dir(self, obj):
        """
        Similar to on_browse, except need to check for valid directory
        :param obj: either in_dir or out_dir TextCtrl objects
        :type obj: wx.TextCtrl
        """
        new_dir = obj.GetValue()
        if isdir(new_dir):
            if obj == self.input['in_dir']:
                self.refresh_ds()
        else:
            ErrorDialog(self, "Please enter a valid directory.", "Directory Error")
            dir_key = 'in' if obj == self.input['in_dir'] else 'out'
            self.directory[dir_key] = new_dir
            if obj == self.input['in_dir']:
                self.file_paths = []
                self.ds = {}
                self.on_parse_complete()

    #################################################################################
    # Combobox Event tickers
    #################################################################################
    def on_file_select(self, *evt):
        self.update_modality()
        self.update_init_value()

    #################################################################################
    # Text Event tickers
    #################################################################################
    def update_dir_obj_text_color(self, evt):
        """Set directory TextCtrl background to orange if directory is invalid"""
        obj = evt.GetEventObject()
        orange = (255, 153, 51, 255)
        color = wx.WHITE if isdir(obj.GetValue()) else orange
        if color != obj.GetBackgroundColour():
            obj.SetBackgroundColour(color)
            self.update_save_dicom_enable()

    #################################################################################
    # Widget Enabling
    #################################################################################
    def update_save_dicom_enable(self):
        """Disable save dicom button if there is not enough information provided"""
        enable = isdir(self.input['in_dir'].GetValue()) and \
                 isdir(self.input['out_dir'].GetValue()) and \
                 self.data_table_has_data
        self.button['save_dicom'].Enable(enable)

    def update_delete_enable(self, *evt):
        """Only enable delete button if edits in the ListCtrl are selected"""
        self.button['delete'].Enable(len(self.data_table.selected_row_data))

    def update_save_template_enable(self):
        """Only enable save button if the ListCtrl has data"""
        enable = self.data_table_has_data
        self.button['save_template'].Enable(enable)
        self.menu_save.Enable(enable)

    #################################################################################
    # Widget updaters
    #################################################################################
    def update_files_found(self):
        """Update the number of files in the GUI"""
        found = len(self.file_paths)
        label = "Files Found: %s" % found
        self.label['files_found'].SetLabel(label)
        self.update_add_enable()

    def update_add_enable(self, *evt):
        enable = len(self.file_paths) > 0 and bool(self.group) and bool(self.element) and self.value_is_valid
        self.button['add'].Enable(enable)
        value_str = self.input['value'].GetValue()
        adv_enable = enable and any([v in value_str for v in ['*fuid[', '*vuid', '*frand[', '*vrand']])
        self.button['advanced'].Enable(adv_enable)
        if enable:
            self.update_preview()
        else:
            self.input['preview'].SetValue('')
        self.update_keyword()

    def update_combobox_files(self):
        """Update the combobox with the file names found in the current in directory"""
        choices = [relpath(f, self.directory['in']) for f in self.file_paths]
        self.input['selected_file'].Enable()
        self.input['selected_file'].SetItems(choices)
        if choices:
            self.input['selected_file'].SetValue(choices[0])

    def update_init_value(self):
        """Update Value in the Tag Editor based on the currently selected file"""
        if self.ds and self.group and self.element:
            ds = self.ds[self.file_paths[self.selected_file]]
            try:
                value = str(ds.get_tag_value(self.tag.tag))
            except Exception:
                try:
                    address = ds.find_tag(self.tag.tag)[0]
                    value = str(address[-1][1])
                except Exception:
                    value = ''
            self.input['value'].SetValue(value)

    def update_modality(self):
        """Update Modality in the Directory box based on the currently selected file"""
        modality = self.ds[self.file_paths[self.selected_file]].modality if self.file_paths else ''
        self.label['modality'].SetLabel('Modality: ' + modality)

    def update_keyword(self):
        """Update Keyword in the Tag Editor based on the current Tag and currently selected file"""
        keyword = self.keyword if self.group and self.element else ''
        self.label['keyword'].SetLabel("Keyword: %s" % keyword)
        self.update_tag_type()
        self.update_vr()

    def update_tag_type(self):
        """Update tag type in the Tag Editor based on the current Tag and currently selected file"""
        tag = self.tag.tag
        tag_type = 'str'
        for file_path in self.file_paths:
            try:
                tag_type = self.ds[file_path].get_tag_type(tag)
                break
            except Exception:
                pass
        self.input['value_type'].SetValue(tag_type)

    def update_vr(self):
        value = self.tag.vr if self.group and self.element else ''
        self.label['value_rep'].SetLabel("Value Representation (VR): %s" % value)

    def update_preview(self):
        """Apply the tag edits to every file in self.ds, return any errors"""
        # TODO: auto-update Value Type based on VR
        # TODO: validate New Value against Value Type
        tag = self.tag
        value_str = self.value
        value_gen = ValueGenerator(value_str, tag.tag, self.current_options)
        file = self.file_paths[self.selected_file]
        values = value_gen(self.ds, file_path=file) if file in self.ds.keys() else ''
        value = str(values[0]) if values else ''
        self.input['preview'].SetValue(value)

    #################################################################################
    # Data updaters
    #################################################################################
    def get_files(self):
        """Get a list of all files in the currently specified in directory"""
        dir_path = self.input['in_dir'].GetValue()
        self.directory['in'] = dir_path
        if isdir(dir_path):
            self.search_sub_folders_last_status = self.search_sub_folders.GetValue()
            self.file_paths = sorted(get_file_paths(dir_path, search_sub_folders=self.search_sub_folders.GetValue()))
        else:
            self.file_paths = []
        self.update_files_found()

    def refresh_ds(self):
        """Update the stored DICOMEditor objects in self.ds"""
        self.get_files()
        self.ds = {}
        self.values_dicts = []
        ParsingProgressFrame(self.file_paths, self.force_open.GetValue())

    #################################################################################
    # Utilities
    #################################################################################
    @property
    def value_is_valid(self):
        value = self.input['value'].GetValue()
        if '*' not in value:  # no functions
            return True
        if value.count('*') % 2 == 1:  # has functions, but missing *
            return False

        call_str = value.split('*')[1::2]

        # each call requires left and right square brackets, an int between them, and a valid function call
        value_functions = [f for f in self.functions if f[0] == 'v']
        for call in call_str:
            if '[' not in call and ']' not in call:
                if call not in value_functions:
                    return False
            else:
                if not (call.count('[') == 1 and call.count(']') == 1 and call.endswith(']')):
                    return False
                if call.split('[')[0] not in self.functions:
                    return False
                param = call.split('[')[1][:-1]
                if not(param.isdigit() or (param.startswith('-') and param[1:].isdigit())):
                    return False

        return True

    @staticmethod
    def path_index(value, n):
        if '*dir[' in value and ']' in value:
            value_split = value.split('*')
            index_temp = value_split[n*2 + 1].split('dir[')[1]
            index_temp_end = index_temp.index(']')
            try:
                return index_temp[:index_temp_end]
            except TypeError:
                pass

    def get_nth_dir_from_file_path(self, value, n, file_path):
        index = self.path_index(value, n)
        if index is not None:
            components = normpath(file_path).split(sep)
            if index < len(components):
                return splitext(components[index])[0], index

    @property
    def dir_contents_have_changed(self):
        current_files = set(get_file_paths(self.directory['in'],
                                           search_sub_folders=self.search_sub_folders_last_status))
        all_files = set(self.file_paths + self.ignored_file_paths)
        return current_files != all_files

    @staticmethod
    def a_referenced_tag_exists(history):
        return any(["Referenced%s" % row[0] in list(keyword_dict) for row in history])

    def get_table_row_data(self, row):
        row_data = self.data_table.get_row(row)
        group = row_data[1].split(',')[0][1:].strip()
        element = row_data[1].split(',')[1][:-1].strip()
        return {'tag': Tag(group, element),
                'keyword': row_data[2],
                'value_str': row_data[3],
                'value_type': get_type(row_data[4]),
                'options': self.all_options[row_data[0]]}

    def set_output_paths(self, check_only=False):
        """
        Save all of the loaded pydicom datasets with the new edits
        :param check_only: If true, perform loop without save to check for overwriting
        :type check_only: bool
        :return: status of file overwriting, if overwrite_check_only=True
        :rtype: bool
        """
        output_dir = self.input['out_dir'].GetValue()
        input_dir = self.directory['in']
        prepend = self.input['prepend_file_name'].GetValue()
        for file_path, ds in self.ds.items():
            file_name = prepend + basename(file_path)
            output_path = join(output_dir, file_name)
            if self.retain_rel_dir.GetValue():
                rel_out_path = join(output_dir, relpath(dirname(file_path), input_dir))
                if not check_only:
                    Path(rel_out_path).mkdir(parents=True, exist_ok=True)
                output_path = join(rel_out_path, file_name)

            if check_only:
                if isfile(output_path):
                    return True
            else:
                ds.output_path = output_path

        if check_only:
            return False

    #################################################################################
    # pub subscribe functions
    #################################################################################
    def add_parsed_data(self, msg):
        self.ds[msg['obj']['dcm']] = msg['data']

    def on_parse_complete(self):

        # remove files that could not be parsed
        bad_files = [(i, f) for i, f in enumerate(self.file_paths) if self.ds[f].dcm is None]
        for (i, f) in bad_files[::-1]:
            self.ignored_file_paths.append(self.file_paths.pop(i))
            self.ds.pop(f)
        self.update_files_found()

        self.update_combobox_files()
        self.update_init_value()
        self.update_modality()

    def add_value_dicts(self, msg):
        self.values_dicts.append(msg['data'])

    def update_dicom_edits(self, msg):
        self.history = msg['data']['history']
        self.error_log = msg['data']['error_log']
        self.ds = msg['data']['ds']

    #################################################################################
    # Finally... initialize and perform dicom edits and save DICOM files
    #################################################################################

    # ------------------------------------------------------------------------------
    # Step 1: Calculate values (time consuming if dynamic value functions used)
    # ------------------------------------------------------------------------------
    def calculate_value_generators(self):
        """Collect a list of Value Generator class objects"""
        self.value_generators = []
        for row in range(self.data_table.row_count):
            row_data = self.get_table_row_data(row)
            self.value_generators.append(ValueGenerator(row_data['value_str'],
                                                        row_data['tag'].tag,
                                                        row_data['options']))
        wx.CallAfter(self.call_next_value_generator)

    def call_next_value_generator(self):
        """Call each Value Generator one at a time, call apply_edits when complete"""
        if self.value_generators:
            value_generator = self.value_generators.pop(0)
            iteration = self.data_table.row_count - len(self.value_generators)
            ValueGenProgressFrame(self.ds, value_generator, iteration, self.data_table.row_count)
        else:
            wx.CallAfter(self.apply_edits)

    # ------------------------------------------------------------------------------
    # Step 2: Perform tag editing
    # ------------------------------------------------------------------------------
    def apply_edits(self):
        """Send table data to apply edits thread, edit DICOM tags"""
        all_row_data = [self.get_table_row_data(row) for row in range(self.data_table.row_count)]
        ApplyEditsProgressFrame(self.ds, self.values_dicts, all_row_data)

    # ------------------------------------------------------------------------------
    # Step 3: View error log, check for file over-writing, sync referenced tags, save DICOM to file
    # ------------------------------------------------------------------------------
    def do_save_dicom(self):
        # If any errors detected, display in GUI and ask user if they'd like to continue
        if self.error_log:
            ViewErrorLog(self.error_log)
            with AskYesNo(self, "Ignore errors and write DICOM files anyway?") as dlg:
                if dlg.ShowModal() == wx.ID_NO:
                    return

        # set_output_paths(check_only=True) will return True if any file would be over-written
        if self.set_output_paths(check_only=True):
            msg = "You will overwrite files with this action. Continue?"
            caption = "Are you sure?"
            flags = wx.ICON_WARNING | wx.YES | wx.NO | wx.NO_DEFAULT
            with wx.MessageDialog(self, msg, caption, flags) as dlg:
                if dlg.ShowModal() == wx.ID_NO:
                    return
        self.set_output_paths()

        # Check for any referenced tags by sending edit history to each data set
        if self.update_referenced_tags.GetSelection() and self.a_referenced_tag_exists(self.history):
            check_all_tags = True if self.update_referenced_tags.GetSelection() == 2 else False
            RefSyncProgressFrame(self.history, self.ds.values(), check_all_tags)
            # This will call SavingProgressFrame when done
        else:
            self.do_saving_progress_frame()

    def do_saving_progress_frame(self):
        SavingProgressFrame(self.ds.values())

    # ------------------------------------------------------------------------------
    # Step 4: Save history, if requested, reparse original data since it has been edited directly
    # ------------------------------------------------------------------------------
    def on_save_complete(self):
        # If in and out directories are the same, need to update file list and datasets with new files
        if self.save_history.GetValue():
            self.save_history_to_file()

        msg = "Re-parse input directory? This is recommended if you wish to apply new edits."
        caption = "Are you sure?"
        flags = wx.ICON_WARNING | wx.YES | wx.NO
        with wx.MessageDialog(self, msg, caption, flags) as dlg:
            if dlg.ShowModal() == wx.ID_NO:
                return
        self.get_files()
        wx.CallAfter(self.refresh_ds)

    def save_history_to_file(self):
        save_file_name = "DVHA_DICOM_Editor_history_%s.csv" % \
                         str(datetime.now()).split('.')[0].replace(':', '-').replace(' ', '-')
        save_file_path = join(self.input['out_dir'].GetValue(), save_file_name)
        column_row = 'File Path,Edit Order,Tag,Old Value,New Value'
        lines = [column_row]
        for file_path, ds in self.ds.items():
            for row in ds.history:
                lines.append('"%s",%s' % (file_path, row))
        with open(save_file_path, 'w') as doc:
            doc.write('\n'.join(lines))

    #################################################################################
    # Help menu events
    #################################################################################
    @staticmethod
    def on_about(*evt):
        About()

    @staticmethod
    def on_githubpage(evt):
        webbrowser.open_new_tab("https://github.com/cutright/DVHA-DICOM-Editor")

    @staticmethod
    def on_report_issue(evt):
        webbrowser.open_new_tab("https://github.com/cutright/DVHA-DICOM-Editor/issues")


class MainApp(wx.App):
    def OnInit(self):
        self.SetAppName('DVHA DICOM Editor')
        self.frame = MainFrame(None, wx.ID_ANY, "DVHA DICOM Editor v%s" % __version__)
        set_frame_icon(self.frame)
        self.SetTopWindow(self.frame)
        self.frame.Show()
        return True


def start():
    app = MainApp(0)
    app.MainLoop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# threads.py
"""
Expensive calculations for main app that merit threading
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

from functools import partial
from dvhaedit.dicom_editor import DICOMEditor, save_dicom, apply_edits, update_referenced_tags
from dvhaedit.threading import ProgressFrame


#############################################################################################
# Simple threads with a single function performed on each item in the provided list
#############################################################################################

class ParsingProgressFrame(ProgressFrame):
    """Create a window to display DICOM file parsing progress and begin ParseWorker"""
    def __init__(self, file_paths, force_open):
        kwargs_list = [{'dcm': f, 'force': force_open} for f in file_paths]
        ProgressFrame.__init__(self, kwargs_list, DICOMEditor, close_msg='parse_complete',
                               action_msg='add_parsed_data', action_gui_phrase='Parsing File',
                               title='Reading DICOM Headers', kwargs=True)


class SavingProgressFrame(ProgressFrame):
    """Create a window to display DICOM file saving progress and begin SaveWorker"""
    def __init__(self, data_sets):
        ProgressFrame.__init__(self, data_sets, save_dicom, close_msg='save_complete',
                               action_gui_phrase='Saving File',
                               title='Saving DICOM Data')


class RefSyncProgressFrame(ProgressFrame):
    """Create a window to display Referenced tag syncing progress and begin SaveWorker"""
    def __init__(self, history, data_sets, check_all_tags):
        ProgressFrame.__init__(self, history, partial(update_referenced_tags, data_sets, check_all_tags),
                               close_msg='ref_sync_complete',
                               action_gui_phrase='Checking References for Tag:',
                               title='Checking for Referenced Tags')


#############################################################################################
# Modified threading implementation
#############################################################################################

# ------------------------------------------------------------------------------
# Value Generator calls GUI progress updates and the provided list is only one item
# ------------------------------------------------------------------------------
class ValueGenProgressFrame(ProgressFrame):
    """Create a window to display value generation progress and begin SaveWorker"""
    def __init__(self, data_sets, value_generator, iteration, total_count):

        ProgressFrame.__init__(self, [data_sets], value_generator,
                               close_msg='value_gen_complete',
                               action_msg='add_value_dicts',
                               action_gui_phrase='File:',
                               title='Generating Values for Tag %s of %s' % (iteration, total_count))


# ------------------------------------------------------------------------------
# Modified to handle nested for-loops in apply_edits
# ------------------------------------------------------------------------------
class ApplyEditsProgressFrame(ProgressFrame):
    """Create a window to display value generation progress and begin SaveWorker"""
    # Input list is only one item, the apply edits function send GUI updates to ProgressFrame
    def __init__(self, data_sets, values_dicts, all_row_data):

        ProgressFrame.__init__(self, [data_sets], partial(apply_edits, values_dicts, all_row_data),
                               close_msg='do_save_dicom',
                               action_msg='update_dicom_edits',
                               action_gui_phrase='',
                               title='Editing DICOM Data')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# paths.py
"""
A collection of directories and paths updated with the script directory
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

from os.path import join, dirname

SCRIPT_DIR = dirname(__file__)
LICENSE_PATH = join(SCRIPT_DIR, 'LICENSE.txt')
DYNAMIC_VALUE_HELP = join(SCRIPT_DIR, 'dynamic_value_help.txt')
ICONS_DIR = join(SCRIPT_DIR, 'icons')
WIN_APP_ICON = join(ICONS_DIR, 'dvha-edit.ico')
WIN_FRAME_ICON = join(ICONS_DIR, 'dvha-edit_frame.ico')
MAC_APP_ICON = join(ICONS_DIR, 'dvha-edit.icns')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# options.py
"""
Classes containing user preferences
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

from dvhaedit.utilities import get_uid_prefixes


class Options:
    def __init__(self):
        # DICOM options
        self.prefix = ''
        self.prefix_dict = get_uid_prefixes()
        self.entropy_source = ''

        # Random Number generator
        self.rand_digits = 5



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dialogs.py
"""
Classes used to edit pydicom datasets
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

import wx
import re
from pydicom.uid import RE_VALID_UID_PREFIX
from dvhaedit._version import __version__
from dvhaedit.data_table import DataTable
from dvhaedit.dicom_editor import TagSearch
from dvhaedit.dynamic_value import HELP_TEXT
from dvhaedit.paths import LICENSE_PATH
from dvhaedit.utilities import save_csv_to_file


class ErrorDialog:
    """This class allows error messages to be called with a one-liner else-where"""

    def __init__(self, parent, message, caption, flags=wx.ICON_ERROR | wx.OK | wx.OK_DEFAULT):
        """
        :param parent: wx parent object
        :param message: error message
        :param caption: error title
        :param flags: flags for wx.MessageDialog
        """
        self.dlg = wx.MessageDialog(parent, message, caption, flags)
        self.dlg.Center()
        self.dlg.ShowModal()
        self.dlg.Destroy()


class AskYesNo(wx.MessageDialog):
    """Simple Yes/No MessageDialog"""

    def __init__(self, parent, msg, caption="Are you sure?",
                 flags=wx.ICON_WARNING | wx.YES | wx.NO | wx.NO_DEFAULT):
        wx.MessageDialog.__init__(self, parent, msg, caption, flags)


class ViewErrorLog(wx.Dialog):
    """Dialog to display the error log in a scrollable window"""
    def __init__(self, error_log):
        """
        :param error_log: error log text
        :type error_log: str
        """
        wx.Dialog.__init__(self, None, title='Error log')

        self.error_log = error_log
        self.button = {'dismiss': wx.Button(self, wx.ID_OK, "Dismiss"),
                       'save': wx.Button(self, wx.ID_ANY, "Save")}
        self.scrolled_window = wx.ScrolledWindow(self, wx.ID_ANY)
        self.text = wx.StaticText(self.scrolled_window, wx.ID_ANY,
                                  "The following errors occurred while editing DICOM tags...\n\n%s" % self.error_log)

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __do_bind(self):
        self.Bind(wx.EVT_BUTTON, self.on_save, id=self.button['save'].GetId())

    def __set_properties(self):
        self.scrolled_window.SetScrollRate(20, 20)
        self.scrolled_window.SetBackgroundColour(wx.WHITE)

    def __do_layout(self):
        # Create sizers
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_text = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)

        # Add error log text
        sizer_text.Add(self.text, 0, wx.EXPAND | wx.ALL, 5)
        self.scrolled_window.SetSizer(sizer_text)
        sizer_wrapper.Add(self.scrolled_window, 1, wx.EXPAND, 0)

        # Add buttons
        sizer_buttons.Add(self.button['save'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_buttons.Add(self.button['dismiss'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_wrapper.Add(sizer_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.SetMinSize((700, 600))
        self.Fit()
        self.Center()

    def run(self):
        """Open dialog, close on Dismiss click"""
        self.ShowModal()
        self.Destroy()

    def on_save(self, *evt):
        """On save button click, create save window to save error log"""
        dlg = wx.FileDialog(self, "Save error log", "", wildcard='*.txt',
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if dlg.ShowModal() == wx.ID_OK:
            save_csv_to_file(self.error_log, dlg.GetPath())
        dlg.Destroy()


class TagSearchDialog(wx.Dialog):
    """A dialog consisting of a search bar and table of partial DICOM Tag matches"""
    def __init__(self, parent):
        """
        :param parent: main frame of DVHA DICOM Edit
        """
        wx.Dialog.__init__(self, parent, title='DICOM Tag Search')

        self.parent = parent

        # Create search bar and TagSearch class
        self.search_ctrl = wx.SearchCtrl(self, wx.ID_ANY, "")
        self.search_ctrl.ShowCancelButton(True)
        self.search = TagSearch()

        self.note = wx.StaticText(self, wx.ID_ANY, "NOTE: The loaded DICOM file(s) may not have the selected tag.")

        # Create table for search results
        columns = ['Keyword', 'Tag', 'VR']
        data = {c: [''] for c in columns}
        self.list_ctrl = wx.ListCtrl(self, wx.ID_ANY, style=wx.BORDER_SUNKEN | wx.LC_REPORT | wx.LC_SINGLE_SEL)
        self.data_table = DataTable(self.list_ctrl, data=data, columns=columns, widths=[-2, -2, -2])

        # Create buttons
        keys = {'select': wx.ID_OK, 'cancel': wx.ID_CANCEL}
        self.button = {key: wx.Button(self, id_, key.capitalize()) for key, id_ in keys.items()}

        self.__do_bind()
        self.__do_layout()

        self.run()

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.update, id=self.search_ctrl.GetId())
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.on_double_click, id=self.list_ctrl.GetId())
        self.Bind(wx.EVT_LIST_COL_CLICK, self.data_table.sort_table, self.list_ctrl)

    def __do_layout(self):
        # Create sizers
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_search = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)

        # Add search bar and results table
        sizer_search.Add(self.search_ctrl, 0, wx.EXPAND | wx.ALL, 5)
        sizer_search.Add(self.note, 0, wx.EXPAND | wx.ALL, 5)
        sizer_search.Add(self.list_ctrl, 1, wx.EXPAND | wx.ALL, 5)
        sizer_main.Add(sizer_search, 1, wx.EXPAND | wx.ALL, 5)

        # Add buttons
        sizer_buttons.Add(self.button['select'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_buttons.Add(self.button['cancel'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        # Add everything to window wrapper
        sizer_wrapper.Add(sizer_main, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.SetMinSize((700, 400))
        self.Fit()
        self.Center()

    def run(self):
        """Open dialog, perform action if Select button is clicked, then close"""
        self.update()
        res = self.ShowModal()
        if res == wx.ID_OK:  # if user clicks Select button
            self.set_tag_to_selection()
        self.Destroy()

    @property
    def data_dict(self):
        """Get the DICOM Tag table data with current search_ctrl value"""
        return self.search(self.search_ctrl.GetValue())

    @property
    def selected_tag(self):
        """Get the Tag of the currently selected/activated row in list_ctrl"""
        selected_data = self.data_table.selected_row_data
        if selected_data:
            return selected_data[0][1]

    def update(self, *evt):
        """Set the table date based on the current search_ctrl value"""
        self.data_table.set_data(**self.data_dict)

    def set_tag_to_selection(self):
        """Set the Group and Element list_ctrl values in the main app"""
        tag = self.selected_tag
        if tag:
            self.parent.input['tag_group'].SetValue(tag.group)
            self.parent.input['tag_element'].SetValue(tag.element)
            self.parent.update_init_value()
            self.parent.update_keyword()

    def on_double_click(self, evt):
        """Treat double-click the same as selecting a row then clicking Select"""
        self.set_tag_to_selection()
        self.Close()


class TextViewer(wx.Dialog):
    """Simple dialog to display the LICENSE file and a brief text header in a scrollable window"""
    def __init__(self, text, title, min_size):
        wx.Dialog.__init__(self, None, title=title)

        self.SetMinSize(min_size)

        self.scrolled_window = wx.ScrolledWindow(self, wx.ID_ANY)
        self.text = wx.StaticText(self.scrolled_window, wx.ID_ANY, text)

        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.scrolled_window.SetScrollRate(20, 20)
        self.SetBackgroundColour(wx.WHITE)

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_text = wx.BoxSizer(wx.VERTICAL)

        sizer_text.Add(self.text, 0, wx.EXPAND | wx.ALL, 5)
        self.scrolled_window.SetSizer(sizer_text)
        sizer_wrapper.Add(self.scrolled_window, 1, wx.EXPAND, 0)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Center()

    def run(self):
        self.ShowModal()
        self.Destroy()


class About(TextViewer):
    """Simple dialog to display the LICENSE file and a brief text header in a scrollable window"""
    def __init__(self):

        with open(LICENSE_PATH, 'r', encoding="utf8") as license_file:
            license_text = ''.join([line for line in license_file])
        license_text = "DVHA DICOM Editor v%s\nedit.dvhanalytics.com\n\n%s" % (__version__, license_text)

        TextViewer.__init__(self, license_text, title='About DVHA DICOM Editor', min_size=(700, 500))


class DynamicValueHelp(TextViewer):
    def __init__(self):
        TextViewer.__init__(self, HELP_TEXT, title='Dynamic Values', min_size=(672, 420))


class AdvancedSettings(wx.Dialog):
    def __init__(self, options):
        wx.Dialog.__init__(self, None, title='User Settings')

        self.options = options

        key_map = {'dicom_prefix': 'Prefix:'}
        self.combo_box = {key: wx.ComboBox(self, wx.ID_ANY, "") for key in key_map.keys()}
        self.label = {key: wx.StaticText(self, wx.ID_ANY, value) for key, value in key_map.items()}

        key_map = {'entropy_source': 'Entropy Source:', 'rand_digits': 'Digits:'}
        self.text_ctrl = {key: wx.TextCtrl(self, wx.ID_ANY, "") for key in key_map.keys()}
        for key, value in key_map.items():
            self.label[key] = wx.StaticText(self, wx.ID_ANY, value)

        self.button = {'ok': wx.Button(self, wx.ID_OK, 'OK'),
                       'cancel': wx.Button(self, wx.ID_CANCEL, 'Cancel')}

        self.valid_prefix_pattern = re.compile(RE_VALID_UID_PREFIX)

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.combo_box['dicom_prefix'].SetItems(sorted(list(self.options.prefix_dict)))
        self.combo_box['dicom_prefix'].SetValue(self.options.prefix)

        self.text_ctrl['entropy_source'].SetValue(self.options.entropy_source)

        self.text_ctrl['rand_digits'].SetValue(str(self.options.rand_digits))

        self.SetMinSize((672, 210))

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.update_ok_enable, id=self.text_ctrl['rand_digits'].GetId())
        self.Bind(wx.EVT_TEXT, self.update_ok_enable, id=self.combo_box['dicom_prefix'].GetId())
        self.Bind(wx.EVT_COMBOBOX, self.update_ok_enable, id=self.combo_box['dicom_prefix'].GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_dicom = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "DICOM UID Generator"), wx.VERTICAL)
        sizer_rand = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Random Number Generator"), wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)

        sizer_dicom.Add(self.label['dicom_prefix'], 0, wx.EXPAND, 0)
        sizer_dicom.Add(self.combo_box['dicom_prefix'], 0, wx.EXPAND, 0)
        sizer_dicom.Add(self.label['entropy_source'], 0, wx.EXPAND, 0)
        sizer_dicom.Add(self.text_ctrl['entropy_source'], 0, wx.EXPAND, 0)
        sizer_main.Add(sizer_dicom, 1, wx.EXPAND, wx.ALL, 5)

        sizer_rand.Add(self.label['rand_digits'], 0, wx.EXPAND, 0)
        sizer_rand.Add(self.text_ctrl['rand_digits'], 0, 0, 0)
        sizer_main.Add(sizer_rand, 0, wx.EXPAND | wx.ALL, 0)  # Has 5 border built-in???

        sizer_buttons.Add(self.button['ok'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_buttons.Add(self.button['cancel'], 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Layout()
        self.Center()

    def run(self):
        if self.ShowModal() == wx.ID_OK:
            self.apply_settings()
        self.Destroy()

    def apply_settings(self):
        self.set_prefix()
        self.set_entropy()
        self.set_rand_digits()

    def set_prefix(self):
        self.options.prefix = self.prefix

    def set_entropy(self):
        self.options.entropy_source = self.text_ctrl['entropy_source'].GetValue()

    def set_rand_digits(self):
        self.options.rand_digits = int(self.text_ctrl['rand_digits'].GetValue())

    def update_ok_enable(self, *evt):
        self.button['ok'].Enable(self.is_input_valid)

    @property
    def is_input_valid(self):
        return self.is_rand_digit_valid and self.is_prefix_valid

    @property
    def is_rand_digit_valid(self):
        value = self.text_ctrl['rand_digits'].GetValue()
        return value.isdigit() and 0 < int(value) <= 64

    @property
    def prefix(self):
        new_value = self.combo_box['dicom_prefix'].GetValue()
        if new_value in self.options.prefix_dict.keys():
            new_value = self.options.prefix_dict[new_value] + '.'
        return new_value

    @property
    def is_prefix_valid(self):
        return self.valid_prefix_pattern.sub('', self.prefix) == ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# utilities.py
"""
General utilities borrowed from DVH Analytics
"""
# Copyright (c) 2016-2020 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from os import walk, listdir
from os.path import isfile, isdir, splitext, join
import pickle
from pydicom._uid_dict import UID_dictionary
import re
from dvhaedit.paths import SCRIPT_DIR, WIN_FRAME_ICON


def get_file_paths(start_path, search_sub_folders=False, extension=None):
    """
    Get a list of absolute file paths for a given directory
    :param start_path: initial directory
    :type start_path str
    :param search_sub_folders: optionally search all sub folders
    :type search_sub_folders: bool
    :param extension: optionally include only files with specified extension
    :type extension: str
    :return: absolute file paths
    :rtype: list
    """

    ignored_files = ['.ds_store']

    if isdir(start_path):
        if search_sub_folders:
            file_paths = []
            for root, dirs, files in walk(start_path, topdown=False):
                for name in files:
                    if extension is None or splitext(name)[1].lower() == extension.lower():
                        if name.lower() not in ignored_files:
                            file_paths.append(join(root, name))
            return file_paths

        file_paths = []
        for f in listdir(start_path):
            if isfile(join(start_path, f)):
                if extension is None or splitext(f)[1].lower() == extension.lower():
                    if f.lower() not in ignored_files:
                        file_paths.append(join(start_path, f))
        return file_paths
    return []


def get_selected_listctrl_items(list_control):
    """
    Get the indices of the currently selected items of a wx.ListCtrl object
    :param list_control: any wx.ListCtrl object
    :type list_control: ListCtrl
    :return: indices of selected items
    :rtype: list
    """
    selection = []

    index_current = -1
    while True:
        index_next = list_control.GetNextItem(index_current, wx.LIST_NEXT_ALL, wx.LIST_STATE_SELECTED)
        if index_next == -1:
            return selection

        selection.append(index_next)
        index_current = index_next


def get_sorted_indices(some_list):
    try:
        return [i[0] for i in sorted(enumerate(some_list), key=lambda x: x[1])]
    except TypeError:  # can't sort if a mix of str and float
        try:
            temp_data = [[value, -float('inf')][value == 'None'] for value in some_list]
            return [i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])]
        except TypeError:
            temp_data = [str(value) for value in some_list]
            return [i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])]


def is_windows():
    return wx.Platform == '__WXMSW__'


def set_msw_background_color(parent):
    if is_windows():
        parent.SetBackgroundColour('lightgrey')


def is_linux():
    return wx.Platform == '__WXGTK__'


def is_mac():
    return wx.Platform == '__WXMAC__'


def get_window_size(width, height):
    """
    Function used to adapt frames/windows for the user's resolution
    :param width: fractional width of the user's screen
    :param height: fractional height of the user's screen
    :return: window size
    :rtype: tuple
    """
    user_width, user_height = wx.GetDisplaySize()
    if user_width / user_height < 1.5:  # catch 4:3 or non-widescreen
        user_height = user_width / 1.6
    return tuple([int(width * user_width), int(height * user_height)])


def get_type(type_str):
    type_map = {'float': float, 'int': int, 'str': str}
    if type_str.lower() in list(type_map):
        return type_map[type_str.lower()]
    return str


def save_csv_to_file(csv_data, abs_file_path):
    """
    Save a python object acceptable for pickle to the provided file path
    """
    with open(abs_file_path, 'w') as outfile:
        outfile.write(csv_data)


def load_csv_from_file(abs_file_path):
    """
    Load a pickled object from the provided absolute file path
    """
    columns, data = None, None
    if isfile(abs_file_path):
        with open(abs_file_path, 'r') as infile:
            columns = [c.strip() for c in infile.readline().split(',')]
            data = {c: [] for c in columns}
            for row in infile:
                for i, value in enumerate(row.split(',')):
                    data[columns[i]].append(value.strip().replace(';', ','))

    return columns, data


def save_object_to_file(obj, abs_file_path):
    """
    Save a python object acceptable for pickle to the provided file path
    """
    with open(abs_file_path, 'wb') as outfile:
        pickle.dump(obj, outfile)


def load_object_from_file(abs_file_path):
    """
    Load a pickled object from the provided absolute file path
    """
    if isfile(abs_file_path):
        with open(abs_file_path, 'rb') as infile:
            obj = pickle.load(infile)
        return obj


def remove_non_alphanumeric(some_string):
    """
    Remove non-alphanumeric characters from a string
    https://stackoverflow.com/questions/1276764/stripping-everything-but-alphanumeric-chars-from-a-string-in-python
    Accessed: Mar 27, 2020
    :param some_string: any string
    :type some_string: str
    :return: the provided string less any non-alphanumeric characters
    """
    pattern = re.compile(r'[\W_]+')
    return pattern.sub('', some_string).replace('_', '')


def remove_characters_between(some_string, lt_char, rt_char):
    regex = lt_char + '[^>]+' + rt_char
    return re.sub(regex, '', some_string)


def update_dynamic_value_help_txt(first_line_value='Dynamic Value Setting'):
    """
    Convert README.md file into plain text.
    Designed to extract text specifically dynamic_value.HELP_TEXT
    :param first_line_value: optionally ignore every line before this value is found
    :type first_line_value: str
    """
    divider = '------------------------------------------------------------------------------'
    new_lines = [divider]
    with open('README.md', 'r') as markdown:
        first_line_found = False
        for line in markdown:
            if line.startswith(first_line_value):
                first_line_found = True
            if first_line_found:
                append_divider = False
                new_line = remove_characters_between(line, '<', '>')
                new_line = new_line.replace('\n', '')  # don't want to strip leading spaces
                if new_line.strip().startswith('* `'):
                    new_line = new_line.replace('* ', '')
                else:
                    if new_line.startswith('### '):
                        new_lines.append(divider)
                        new_line = new_line[4:].upper()
                        append_divider = True
                    new_line = new_line.replace('* ', '').replace('*', '')

                new_line = new_line.replace(' ' * 4, ' ' * 8)  # double the indents
                new_line = new_line.replace('`', '')
                new_lines.append(new_line)
                if append_divider:
                    new_lines.append(divider)

    with open(join(SCRIPT_DIR, 'dynamic_value_help.txt'), 'w') as doc:
        doc.write('\n'.join(new_lines))


def set_frame_icon(frame):
    if not is_mac():
        frame.SetIcon(wx.Icon(WIN_FRAME_ICON))


def get_uid_prefixes():
    prefix_dict = {}
    for prefix, data in UID_dictionary.items():
        if data[0]:
            key = "%s - %s" % (data[0], data[1])
            if data[3].lower() == 'retired':
                key = key + ' (Retired)'
            prefix_dict[key] = prefix
    return prefix_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dvhaedit_app.py
"""
Script to start DVHA DICOM Editor
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

import dvhaedit.main

if __name__ == "__main__":
    dvhaedit.main.start()
